#include "stdafx.h"

#define BLOCK_SIZE  4608
#define BLOCK_COUNT 20

HWAVEOUT hWaveOut = NULL;

static void CALLBACK waveOutProc(HWAVEOUT, UINT, DWORD, DWORD, DWORD);
static WAVEHDR* allocateBlocks(int size, int count);
static void freeBlocks(WAVEHDR* blockArray);
static void writeAudio(HWAVEOUT hWaveOut, LPSTR data, int size);

static CRITICAL_SECTION waveCriticalSection;
static WAVEHDR*         waveBlocks;
static volatile unsigned int     waveFreeBlockCount;
static int              waveCurrentBlock;

typedef struct AudioState {
	AVFormatContext *pFmtCtx;
	AVCodecContext *pCodecCtx;
	AVCodec* pCodec;
	uint8_t audio_buf1[(AVCODEC_MAX_AUDIO_FRAME_SIZE * 3) / 2];
	uint8_t *audio_buf;
	unsigned int audio_buf_size; /* in bytes */
	int audio_buf_index; /* in bytes */
	AVPacket audio_pkt_temp;
	AVPacket audio_pkt;
	uint8_t *audio_pkt_data;
	int audio_pkt_size;

	int stream_index;
} AudioState;

int audio_decode_frame(AudioState* pState) {

	AVPacket* pkt_temp = &pState->audio_pkt_temp;
	AVPacket* pkt = &pState->audio_pkt;
	AVCodecContext* dec= pState->pCodecCtx;
	int len = 0, data_size = sizeof(pState->audio_buf1);
	int err = 0;

	for( ; ; ) {
		while (pkt_temp->size > 0) {
			data_size = sizeof(pState->audio_buf1);
			len = avcodec_decode_audio3(dec, (int16_t*)pState->audio_buf1, &data_size, pkt_temp);
			if (len < 0) {
				/* if error, we skip the frame */
				pkt_temp->size = 0;
				break;
			}

			pkt_temp->data += len;
			pkt_temp->size -= len;

			if (data_size <= 0)
				continue;

			pState->audio_buf = pState->audio_buf1;
			return data_size;
		}

		if (pkt->data)
			av_free_packet(pkt);

		if((err = av_read_frame(pState->pFmtCtx, pkt)) < 0)
			return -1;

		pkt_temp->data = pkt->data;
		pkt_temp->size = pkt->size;
	}

	return -1;
}

#define INBUF_SIZE 4096

int _tmain0(int argc, _TCHAR* argv[])
{
	AVCodec *codec;
	AVCodecContext *c= NULL;
	int out_size, len;
	FILE *f;
	uint8_t *outbuf;
	uint8_t inbuf[INBUF_SIZE + FF_INPUT_BUFFER_PADDING_SIZE];
	AVPacket avpkt;
	char filename[] = "d:\\music\\.aac";
	unsigned int i = 0;

	WAVEFORMATEX wfx = {0};
	wfx.nSamplesPerSec  = 44100;
	wfx.wBitsPerSample = 16;
	wfx.nChannels       = 2;
	wfx.cbSize          = 0;
	wfx.wFormatTag      = WAVE_FORMAT_PCM;
	wfx.nBlockAlign     = (wfx.wBitsPerSample * wfx.nChannels) >> 3;
	wfx.nAvgBytesPerSec = wfx.nBlockAlign * wfx.nSamplesPerSec;

	av_register_all();
	av_init_packet(&avpkt);

	/* find the mpeg audio decoder */
	codec = avcodec_find_decoder(CODEC_ID_AAC);
	if (!codec) {
		fprintf(stderr, "codec not found\n");
		exit(1);
	}

	c = avcodec_alloc_context();

	/* open it */
	if (avcodec_open(c, codec) < 0) {
		fprintf(stderr, "could not open codec\n");
		exit(1);
	}

	outbuf = (uint8_t*)malloc(AVCODEC_MAX_AUDIO_FRAME_SIZE);

	f = fopen(filename, "rb");
	if (!f) {
		fprintf(stderr, "could not open %s\n", filename);
		exit(1);
	}

	waveBlocks         = allocateBlocks(BLOCK_SIZE, BLOCK_COUNT);
	waveFreeBlockCount = BLOCK_COUNT;
	waveCurrentBlock   = 0;

	InitializeCriticalSection(&waveCriticalSection);

	// open wave out device
	if(waveOutOpen(&hWaveOut, WAVE_MAPPER,  &wfx, (DWORD_PTR)waveOutProc, 
		(DWORD_PTR)&waveFreeBlockCount, CALLBACK_FUNCTION) != MMSYSERR_NOERROR) {
			fprintf(stderr, "%s: unable to open wave mapper device\n", argv[0]);
			ExitProcess(1);
	}

	memset(inbuf, 0, sizeof(inbuf));
	memset(outbuf, 0, AVCODEC_MAX_AUDIO_FRAME_SIZE);

	avpkt.data = inbuf;
	for( ; ; ) {
		avpkt.size = fread(inbuf, 1, INBUF_SIZE, f);
		if (avpkt.size == 0)
			break;

		avpkt.data = inbuf;
		while (avpkt.size > 0) {
			out_size = AVCODEC_MAX_AUDIO_FRAME_SIZE;
			len = avcodec_decode_audio3(c, (short *)outbuf, &out_size, &avpkt);
			if (len < 0) {
				fprintf(stderr, "Error while decoding\n");
				break;
			}

			if (out_size > 0) {
				/* if a frame has been decoded, output it */
				writeAudio(hWaveOut, (char*)outbuf, out_size);
			}

			avpkt.size -= len;
			avpkt.data += len;
		}
	}

	// wait for complete
	for( ; ; ) {
		if(waveFreeBlockCount >= BLOCK_COUNT)
			break;

		Sleep(10);
	}

	for(i = 0; i < waveFreeBlockCount; i++)
		if(waveBlocks[i].dwFlags & WHDR_PREPARED)
			waveOutUnprepareHeader(hWaveOut, &waveBlocks[i], sizeof(WAVEHDR));

	DeleteCriticalSection(&waveCriticalSection);
	freeBlocks(waveBlocks);
	waveOutClose(hWaveOut);

	fclose(f);
	free(outbuf);

	avcodec_close(c);
	av_free(c);

	return 0;
}

int _tmain(int argc, _TCHAR* argv[]) {
	int err = 0;
	AudioState audio_state = {0};
	unsigned int i = 0;
	unsigned int ready = 0;

	OPENFILENAME ofn = {0};
	TCHAR filename[MAX_PATH];
	WAVEFORMATEX wfx = {0};
	uint8_t buffer[BLOCK_SIZE];
	uint8_t* pbuffer = buffer;
	AVInputFormat* iformat = NULL;
	
	int audio_size = 0, data_size = 0;
	int len = 0, len1 = 0, eof = 0, size = 0;

	memset(&ofn, 0, sizeof(OPENFILENAME));
	ofn.lStructSize = sizeof(ofn);
	ofn.hwndOwner = GetDesktopWindow();
	ofn.lpstrFile = filename;
	ofn.lpstrFile[0] = _TEXT('\0');
	ofn.nMaxFile = sizeof(filename) / sizeof(filename[0]);
	ofn.lpstrFilter = _TEXT("All support files\0*.aac;*.ape;*.flac;*.mp3;*.mp4;*.mpc;*.ogg;*.tta;*.wma;*.wav\0");
	ofn.nFilterIndex = 1;
	ofn.lpstrFileTitle = NULL;
	ofn.nMaxFileTitle = 0;
	ofn.lpstrInitialDir = NULL;
	ofn.Flags = OFN_PATHMUSTEXIST | OFN_FILEMUSTEXIST;

	if (GetOpenFileName(&ofn) == FALSE)
		return 0;

	av_register_all();

	//iformat = av_iformat_next(iformat);
	//while(iformat)
	//{
	//	printf("%s[%s]\n", iformat->name, iformat->long_name);
	//	iformat = iformat->next;
	//}

	err = av_open_input_file(&audio_state.pFmtCtx, filename, NULL, 0, NULL);
	if(err < 0) {
		printf("can not open file %s.\n", filename);
		return -1;
	}

	err = av_find_stream_info(audio_state.pFmtCtx);
	if(err < 0) {
		printf("can not find stream info of file %s.\n", filename);
		return -1;
	}

	for(i = 0; i < audio_state.pFmtCtx->nb_streams; i++) {
		if(audio_state.pFmtCtx->streams[i]->codec->codec_type == CODEC_TYPE_AUDIO) {
			audio_state.pCodecCtx = audio_state.pFmtCtx->streams[i]->codec;
			audio_state.stream_index = i;
			ready = 1;
			break;
		}
	}

	if(!ready)
		return -1;

	printf("codec_id=%d\n", audio_state.pCodecCtx->codec_id);

	audio_state.pCodec = avcodec_find_decoder(audio_state.pCodecCtx->codec_id);
	if(!audio_state.pCodec || avcodec_open(audio_state.pCodecCtx, audio_state.pCodec) < 0)
		return -1;

	wfx.nSamplesPerSec  = audio_state.pCodecCtx->sample_rate;
	switch(audio_state.pCodecCtx->sample_fmt)
	{
	case SAMPLE_FMT_U8:
		wfx.wBitsPerSample = 8;
		break;
	case SAMPLE_FMT_S16:
		wfx.wBitsPerSample = 16;
		break;
	case SAMPLE_FMT_S32:
		wfx.wBitsPerSample = 32;
		break;
	case SAMPLE_FMT_FLT:
		wfx.wBitsPerSample = sizeof(double) * 8;
		break;
	default:
		wfx.wBitsPerSample = 0;
		break;
	}

	wfx.nChannels       = FFMIN(2, audio_state.pCodecCtx->channels);
	wfx.cbSize          = 0;
	wfx.wFormatTag      = WAVE_FORMAT_PCM;
	wfx.nBlockAlign     = (wfx.wBitsPerSample * wfx.nChannels) >> 3;
	wfx.nAvgBytesPerSec = wfx.nBlockAlign * wfx.nSamplesPerSec;

	printf("sample rate: %d\n", wfx.nSamplesPerSec);
	printf("channel count: %d\n", wfx.nChannels);
	printf("bit per sample: %d\n", wfx.wBitsPerSample);

	waveBlocks         = allocateBlocks(BLOCK_SIZE, BLOCK_COUNT);
	waveFreeBlockCount = BLOCK_COUNT;
	waveCurrentBlock   = 0;

	InitializeCriticalSection(&waveCriticalSection);

	// open wave out device
	if(waveOutOpen(&hWaveOut, WAVE_MAPPER,  &wfx, (DWORD_PTR)waveOutProc, 
		(DWORD_PTR)&waveFreeBlockCount, CALLBACK_FUNCTION) != MMSYSERR_NOERROR) {
		fprintf(stderr, "%s: unable to open wave mapper device\n", argv[0]);
		ExitProcess(1);
	}

	// memset(&audio_state.audio_pkt, 0, sizeof(AVPacket));
	// memset(&audio_state.audio_pkt_temp, 0, sizeof(AVPacket));
	// av_init_packet(&audio_state.audio_pkt);

	// play loop
	for( ; ; ) {

		len = BLOCK_SIZE;
		size = 0;
		pbuffer = buffer;

		if(eof)
			break;

		while(len > 0) {
			if(audio_state.audio_buf_index >= (int)audio_state.audio_buf_size) {
				audio_size = audio_decode_frame(&audio_state);
				if(audio_size < 0) {
					if(size > 0)
						break;

					eof = 1;
					break;
				}

				audio_state.audio_buf_size = audio_size;
				audio_state.audio_buf_index = 0;
			}

			len1 = audio_state.audio_buf_size - audio_state.audio_buf_index;
			if(len1 > len)
				len1 = len;

			memcpy(pbuffer, (uint8_t *)audio_state.audio_buf + audio_state.audio_buf_index, len1);

			len -= len1;
			pbuffer += len1;
			size += len1;
			audio_state.audio_buf_index += len1;
		}

		printf("size=%d\n", size);
		writeAudio(hWaveOut, (char*)buffer, size);
	}

	// wait for complete
	for( ; ; ) {
		if(waveFreeBlockCount >= BLOCK_COUNT)
			break;

		Sleep(10);
	}

	for(i = 0; i < waveFreeBlockCount; i++)
		if(waveBlocks[i].dwFlags & WHDR_PREPARED)
			waveOutUnprepareHeader(hWaveOut, &waveBlocks[i], sizeof(WAVEHDR));

	DeleteCriticalSection(&waveCriticalSection);
	freeBlocks(waveBlocks);
	waveOutClose(hWaveOut);

	avcodec_close(audio_state.pCodecCtx);

	system("pause");
	return 0;
}

static void writeAudio(HWAVEOUT hWaveOut, LPSTR data, int size)
{
	WAVEHDR* current;
	int remain;

	current = &waveBlocks[waveCurrentBlock];

	while(size > 0) {
		/* 
		* first make sure the header we're going to use is unprepared
		*/
		if(current->dwFlags & WHDR_PREPARED)
			waveOutUnprepareHeader(hWaveOut, current, sizeof(WAVEHDR));

		if(size < (int)(BLOCK_SIZE - current->dwUser)) {
			memcpy(current->lpData + current->dwUser, data, size);
			current->dwUser += size;
			break;
		}

		remain = BLOCK_SIZE - current->dwUser;
		memcpy(current->lpData + current->dwUser, data, remain);
		size -= remain;
		data += remain;
		current->dwBufferLength = BLOCK_SIZE;

		waveOutPrepareHeader(hWaveOut, current, sizeof(WAVEHDR));
		waveOutWrite(hWaveOut, current, sizeof(WAVEHDR));

		EnterCriticalSection(&waveCriticalSection);
		waveFreeBlockCount--;
		LeaveCriticalSection(&waveCriticalSection);

		/*
		* wait for a block to become free
		*/
		while(!waveFreeBlockCount)
			Sleep(10);

		/*
		* point to the next block
		*/
		waveCurrentBlock++;
		waveCurrentBlock %= BLOCK_COUNT;

		current = &waveBlocks[waveCurrentBlock];
		current->dwUser = 0;
	}
}

static WAVEHDR* allocateBlocks(int size, int count)
{
	char* buffer;
	int i;
	WAVEHDR* blocks;
	DWORD totalBufferSize = (size + sizeof(WAVEHDR)) * count;

	/*
	* allocate memory for the entire set in one go
	*/
	if((buffer = (char*)HeapAlloc(
		GetProcessHeap(), 
		HEAP_ZERO_MEMORY, 
		totalBufferSize
		)) == NULL) {
			fprintf(stderr, "Memory allocation error\n");
			ExitProcess(1);
	}

	/*
	* and set up the pointers to each bit
	*/
	blocks = (WAVEHDR*)buffer;
	buffer += sizeof(WAVEHDR) * count;
	for(i = 0; i < count; i++) {
		blocks[i].dwBufferLength = size;
		blocks[i].lpData = buffer;
		buffer += size;
	}

	return blocks;
}

static void freeBlocks(WAVEHDR* blockArray)
{
	/* 
	* and this is why allocateBlocks works the way it does
	*/ 
	HeapFree(GetProcessHeap(), 0, blockArray);
}

static void CALLBACK waveOutProc(
								 HWAVEOUT hWaveOut, 
								 UINT uMsg, 
								 DWORD dwInstance,  
								 DWORD dwParam1,    
								 DWORD dwParam2     
								 )
{
	int* freeBlockCounter = (int*)dwInstance;
	/*
	* ignore calls that occur due to opening and closing the
	* device.
	*/
	if(uMsg != WOM_DONE)
		return;

	EnterCriticalSection(&waveCriticalSection);
	(*freeBlockCounter)++;
	LeaveCriticalSection(&waveCriticalSection);
}

